/*    This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
 *    See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
 *    Author(s):       [AUTHOR NAME]
 *
 *    Copyright (C) [YEAR] [COPYRIGHT]
 *
 *    Modification(s):
 *      - YYYY/MM Author: Description of the modification
 */

#ifndef [FILE_NAME]_H_
#define [FILE_NAME]_H_

#endif  // [FILE_NAME]_H_